/* Big FAT File Recoverer */

#include <stdio.h>
#include <string.h>

unsigned int get2b(char *p) { return p[0]&0xff | (p[1]&0xff)<<8; }
unsigned int get4b(char *p) { return get2b(&p[2])<<16 | get2b(&p[0]); }

void revstr(char *s, int l) {
  int i;

  for (i = 0; i < l/2; i++) {
    char tmp = s[i];
    s[i] = s[l - i - 1];
    s[l - i - 1] = tmp;
  }
}


int get_time(int *h, int *m, int *s, char *ent) {
  unsigned int hms_val = get2b(ent + 0x0e);

  *h = hms_val >> 11;
  *m = (hms_val >> 5)&0x3f;
  *s = (hms_val & 0x1f)*2;

  return *h < 24 && *m < 60 && *s < 60;
}

int get_date(int *y, int *m, int *d, char *ent) {
  unsigned int ymd_val = get2b(ent + 0x10);

  *y = (ymd_val >> 9) + 1980;
  *m = (ymd_val >> 5) & 0xf;
  *d = ymd_val & 0x1f;

  return *y < 2016 && *m > 0 && *m < 13 && *d < 32; 
}

int main(int argc, char **argv) {
  FILE *f = fopen(argv[1], "r");

  // Empirically-discovered offset at which filesystem started
  unsigned offset = 0x1d40, clustersize = 32;
  fseek(f, offset * 512, SEEK_SET);
  
  int i, lfn_idx = 0;
  char sector[512], lfn_text[512];
  unsigned long sector_idx = 0, idx;
  while (fread(sector, 1, 512, f)) {
    int announced_sector = 0;
    int chunk, candidate;
    idx = sector_idx;
    for (chunk = 0; chunk < 512/32; chunk++) {
      char *a = &sector[chunk*32];
      int attrib = a[0x0b], subdir = attrib &0x10, lfn = (attrib == 0x0f),
          size = get4b(&a[0x1c]);

      candidate = 1;
      
      for (i = 0; i < 11; i++) {
        if (!(a[i] == ' ' || a[i] == '~' || a[i] == '_' || isalnum(a[i])))
          candidate = 0;
        /* if (a[i] <= 'z' && a[i] >= 'a') candidate = 0; */
      }

      /* Broken attributes => not a directory */
      if (attrib & 0xc0) candidate = 0;
      
      if (a[0x0d] > 199) candidate = 0;

      int h, m, s, yy, mm, dd;
      if (!get_time(&h, &m, &s, a)) candidate = 0;
      if (!get_date(&yy, &mm, &dd, a)) candidate = 0;

      if (yy < 2010) candidate = 0;

      unsigned int startsat = get2b(&a[0x14])<<16 | get2b(&a[0x1a]);
      if (startsat & 0xff000000) candidate = 0;
      
      if (candidate && !lfn) {
        if (!announced_sector) {
          announced_sector = 1;
          printf("=== SECTOR %08x ===\n", sector_idx/512); 
        }

        lfn_text[lfn_idx] = 0;
        revstr(lfn_text, lfn_idx);
        if (lfn_text[0]) {
          printf("LFN: %s\n", lfn_text);
        } else {
          int t = 0;
          for (i = 0; i < 8 && a[i] != ' '; ++i) lfn_text[t++] = a[i];
          if (a[8] != ' ') lfn_text[t++] = '.';
          for (i = 0; i < 3; ++i) lfn_text[t++] = a[i + 8];
          lfn_text[t] = 0;
        }
        lfn_idx = 0;
        
        printf("Candidate at 0x%09lx: ", idx);
        for (i = 0; i < 11; i++) putc(a[i], stdout);
        if (subdir) printf(" [DIR]");
        printf(" attrib=%02x\t%04d-%02d-%02d", attrib, yy, mm, dd);
        putc('\n', stdout);
        printf("Starts at cluster %08x; %d bytes\n", startsat, size);
        printf(" $ dd if=%s of=\"",argv[1]);
        printf("%s", lfn_text);
        printf("\" bs=512 skip=%d count=%d\n",
               offset + startsat*clustersize, (size + 511)/512);
               
        if (subdir) {
          printf(" # %d %d \"%s\"\n",
            sector_idx/512/clustersize, startsat, lfn_text);
        } else {
          printf(" @ %d %d %d \"%s\"\n",
                 sector_idx/512/clustersize, startsat, size, lfn_text);
        }
               
        for (i = 0; i < 32; i++) {
          printf(" %02x", a[i]&0xff);
          if (i%16 == 15) putc('\n', stdout);
        }
      } else if (lfn) {
        int seq = a[0]&0x1f,
            invalid = !a[0] || a[0]&0x20 || a[0x0c] || get2b(&a[0x1a]);
        if (a[0]&0x40) lfn_idx = 0;
        if (invalid) {
          lfn = 0;
        } else {
          for (i = 12; i >= 0; i--) {
            int idx, val;
            if (i < 5) idx = 0x01 + i*2;
            else if (i < 11) idx = 0x0e + (i - 5)*2;
            else idx = 0x1c + (i - 11)*2;
            lfn_text[lfn_idx] = a[idx];
            lfn_idx++;

            if (lfn_idx >= 511) lfn = lfn_idx = 0;
          }
        }
      }

      if (!(candidate || lfn)) continue;
      
      idx += 32;
    }
    sector_idx += 512;
  }
  
  return 0;
}
